@@ -11,7 +11,6 @@ use File::Spec;
 
 use lib '.';
 use build::setup;
-use build::auto;
 use build::probe;
 
 # This allows us to run on ancient perls.
@@ -41,7 +40,7 @@ my @args = @ARGV;
 
 GetOptions(\%args, qw(
     help|?
-    debug:s optimize:s instrument! coverage
+    debug:s optimize:s coverage
     os=s shell=s toolchain=s compiler=s
     ar=s cc=s ld=s make=s has-sha has-libuv
     static has-libtommath has-libatomic_ops
@@ -92,7 +91,7 @@ if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
     ".\n    Leave off the --relocatable flag to do a non-relocatable build.");
 }
 
-for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
+for (qw(coverage static big-endian has-libtommath has-sha has-libuv
         has-libatomic_ops asan ubsan tsan valgrind dtrace show-vec)) {
     $args{$_} = 0 unless defined $args{$_};
 }
@@ -217,7 +216,7 @@ if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
 
 # Disable RETGUARD on OpenBSD, since it breaks the legojit.
 if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
-	$config{ccmiscflags} .= ' -fno-ret-protector';
+    $config{ccmiscflags} .= ' -fno-ret-protector';
 }
 
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
@@ -365,7 +364,9 @@ else {
                         . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n";
 }
 
-if ($config{pkgconfig_works} && has_native_library('libzstd')) {
+# The ZSTD_CStream API is only exposed starting at version 1.0.0
+# Ubuntu 16.04 packages version 0.5.1, so an "exists" check is not good enough.
+if ($config{pkgconfig_works} && has_native_library('libzstd', '1.0.0')) {
     setup_native_library('libzstd');
     $config{heapsnapformat} = 3;
 }
@@ -397,8 +398,7 @@ my @cflags;
 push @cflags, $config{ccmiscflags};
 push @cflags, $config{ccoptiflags}  if $args{optimize};
 push @cflags, $config{ccdebugflags} if $args{debug};
-push @cflags, $config{ccinstflags}  if $args{instrument};
-push @cflags, $config{ld_covflags}  if $args{coverage};
+push @cflags, $config{cc_covflags}  if $args{coverage};
 push @cflags, $config{ccwarnflags};
 push @cflags, $config{ccdefflags};
 push @cflags, $config{ccshared}     unless $args{static};
@@ -438,7 +438,6 @@ $config{cflags} = join ' ', uniq(@cflags);
 my @ldflags = ($config{ldmiscflags});
 push @ldflags, $config{ldoptiflags}  if $args{optimize};
 push @ldflags, $config{lddebugflags} if $args{debug};
-push @ldflags, $config{ldinstflags}  if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
 if (not $args{static} and $config{prefix} ne '/usr') {
     push @ldflags, $config{ldrpath_relocatable} if  $args{relocatable};
@@ -487,23 +486,52 @@ $config{auxclean} = @auxfiles ? '$(RM) ' . join ' ', @auxfiles : '@:';
 
 print "OK\n\n";
 
-if ($config{crossconf}) {
-    build::auto::detect_cross(\%config, \%defaults);
-    build::probe::static_inline_cross(\%config, \%defaults);
-    build::probe::thread_local_cross(\%config, \%defaults);
-    build::probe::substandard_pow_cross(\%config, \%defaults);
-    build::probe::unaligned_access_cross(\%config, \%defaults);
-    build::probe::ptr_size_cross(\%config, \%defaults);
-}
-else {
-    build::auto::detect_native(\%config, \%defaults);
-    build::probe::static_inline_native(\%config, \%defaults);
-    build::probe::thread_local_native(\%config, \%defaults);
-    build::probe::substandard_pow(\%config, \%defaults);
-    build::probe::unaligned_access(\%config, \%defaults);
-    build::probe::ptr_size_native(\%config, \%defaults);
+unless ($config{crossconf}) {
+    # detect x64 on Windows so we can build the correct dyncall version
+    if ($config{cc} eq 'cl') {
+        print dots('    auto-detecting x64 toolchain');
+        my $msg = `cl 2>&1`;
+        if (defined $msg) {
+            if ($msg =~ /x64/) {
+                print "YES\n";
+                $defaults{-thirdparty}->{dc}->{rule} =
+                    'cd 3rdparty/dyncall && configure.bat /target-x64 && $(MAKE) -f Nmakefile';
+            }
+            else { print "NO\n" }
+        }
+        else {
+            softfail("could not run 'cl'");
+            print dots('    assuming x86'), "OK\n";
+        }
+    }
+    elsif ($defaults{os} eq 'mingw32' && $defaults{-toolchain} eq 'gnu') {
+        print dots('    auto-detecting x64 toolchain');
+        my $cc = $config{cc};
+        my $msg = `$cc -dumpmachine 2>&1`;
+        if (defined $msg) {
+            if ($msg =~ /x86_64/) {
+                print "YES\n";
+
+                $defaults{-thirdparty}->{dc}->{rule} =
+                    'cd 3rdparty/dyncall && ./configure.bat /target-x64 /tool-gcc && $(MAKE) COMPILE.C=$$(COMPILE.c) -f Makefile.embedded mingw32';
+            }
+            else { print "NO\n" }
+        }
+        else {
+            softfail("could not run 'cl'");
+            print dots('    assuming x86'), "OK\n";
+        }
+    }
 }
 
+build::probe::static_inline(\%config, \%defaults);
+build::probe::thread_local(\%config, \%defaults);
+build::probe::substandard_pow(\%config, \%defaults);
+build::probe::substandard_log(\%config, \%defaults);
+build::probe::substandard_trig(\%config, \%defaults);
+build::probe::has_isinf_and_isnan(\%config, \%defaults);
+build::probe::unaligned_access(\%config, \%defaults);
+build::probe::ptr_size(\%config, \%defaults);
 
 my $archname = $Config{archname};
 if ($args{'jit'}) {
@@ -555,16 +583,15 @@ if ($config{cc} eq 'cl') {
                         . "\t\$(CP) 3rdparty/msinttypes/*.h \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n";
 }
 
+if ($^O eq 'aix' && $config{ptr_size} == 4) {
+    $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000');
+}
+
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
 build::probe::pthread_setname_np(\%config, \%defaults);
 build::probe::check_fn_malloc_trim(\%config, \%defaults);
-if ($^O eq 'aix') {
-    build::probe::numbits(\%config, \%defaults);
-    $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000')
-        if $config{arch_bits} == 32;
-}
 build::probe::rdtscp(\%config, \%defaults);
 
 my $order = $config{be} ? 'big endian' : 'little endian';
@@ -711,8 +738,20 @@ sub setup_native {
     print dots("Configuring native build environment");
     print "\n";
 
-    $os = build::probe::win32_compiler_toolchain(\%config, \%defaults)
-        if $os eq 'MSWin32';
+    if ($os eq 'MSWin32') {
+        my $has_nmake = 0 == system('nmake /? >NUL 2>&1');
+        my $has_cl    = `cl 2>&1` =~ /Microsoft Corporation/;
+        my $has_gmake = 0 == system('gmake --version >NUL 2>&1');
+        my $has_gcc   = 0 == system('gcc --version >NUL 2>&1');
+        if ($has_nmake && $has_cl) {
+            $os = 'win32';
+        }
+        elsif ($has_gmake && $has_gcc) {
+            $os = 'mingw32';
+        } else {
+            $os = "";
+        }
+    }
 
     if (!exists $::SYSTEMS{$os}) {
         softfail("unknown OS '$os'");
@@ -838,7 +877,18 @@ sub configure {
                 join(', ', sort keys %config));
         }
 
-        $template =~ s/@\Q$key\E@/$config{$key}/;
+        my $val = $config{$key};
+        if ($template =~ /'[^']*@\Q$key\E[^']*'/) {
+            # escape \ and '
+            $val =~ s/\\/\\\\/g;
+            $val =~ s/\'/\\\'/g;
+        } elsif ($template =~ /"[^"]*@\Q$key\E[^"]*"/) {
+            # escape \ and "
+            $val =~ s/\\/\\\\/g;
+            $val =~ s/\"/\\\"/g;
+        }
+
+        $template =~ s/@\Q$key\E@/$val/;
     }
 
     return $template;
@@ -960,14 +1010,11 @@ sub setup_native_library {
 }
 
 sub has_native_library {
-    my $library = shift;
-    my $result_exists = `$config{pkgconfig} --exists $library`;
-    if ($? == 0) {
-        return 1;
-    }
-    else {
-        return 0;
+    my ($library, $version) = @_;
+    if (defined $version) {
+        return !system "$config{pkgconfig} --atleast-version=$version $library";
     }
+    return !system "$config{pkgconfig} --exists $library"
 }
 
 __END__
@@ -979,7 +1026,7 @@ __END__
     ./Configure.pl [--os <os>] [--shell <shell>]
                    [--toolchain <toolchain>] [--compiler <compiler>]
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
-                   [--debug] [--optimize] [--instrument]
+                   [--debug] [--optimize]
                    [--static] [--prefix <path>] [--relocatable]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
@@ -988,7 +1035,7 @@ __END__
 
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
-                   [--debug] [--optimize] [--instrument]
+                   [--debug] [--optimize]
                    [--static] [--prefix <path>] [--relocatable]
                    [--big-endian] [--make-install]
 
@@ -1022,13 +1069,6 @@ default.
 Toggle optimization and debug flags during compile and link. If nothing
 is specified the default is to optimize.
 
-=item --instrument
-
-=item --no-instrument
-
-Toggle extra instrumentation flags during compile and link; for example,
-turns on Address Sanitizer when compiling with C<clang>.  Defaults to off.
-
 =item --os <os>
 
 Set the operating system name which you are compiling to.
