@@ -13,13 +13,26 @@
 #define DEFAULT_MODE _S_IWRITE /* work around sucky libuv defaults */
 #endif
 
+static MVMint64 file_info_with_error(MVMThreadContext *tc, uv_stat_t* stat, MVMString *filename, MVMint32 use_lstat) {
+    char * const a = MVM_string_utf8_c8_encode_C_string(tc, filename);
+    uv_fs_t req;
+
+    MVMint64 res = use_lstat
+      ? uv_fs_lstat(NULL, &req, a, NULL)
+      :  uv_fs_stat(NULL, &req, a, NULL);
+    *stat = req.statbuf;
+
+    MVM_free(a);
+    return res;
+}
+
 static uv_stat_t file_info(MVMThreadContext *tc, MVMString *filename, MVMint32 use_lstat) {
     char * const a = MVM_string_utf8_c8_encode_C_string(tc, filename);
     uv_fs_t req;
 
     if ((use_lstat
-      ? uv_fs_lstat(tc->loop, &req, a, NULL)
-      :  uv_fs_stat(tc->loop, &req, a, NULL)
+      ? uv_fs_lstat(NULL, &req, a, NULL)
+      :  uv_fs_stat(NULL, &req, a, NULL)
     ) < 0) {
         MVM_free(a);
         MVM_exception_throw_adhoc(tc, "Failed to stat file: %s", uv_strerror(req.result));
@@ -41,8 +54,8 @@ MVMint64 MVM_file_stat(MVMThreadContext *tc, MVMString *filename, MVMint64 statu
                 uv_fs_t req;
 
                 if ((use_lstat
-                  ? uv_fs_lstat(tc->loop, &req, a, NULL)
-                  :  uv_fs_stat(tc->loop, &req, a, NULL)
+                  ? uv_fs_lstat(NULL, &req, a, NULL)
+                  :  uv_fs_stat(NULL, &req, a, NULL)
                 ) < 0) {
                     MVM_free(a);
                     MVM_exception_throw_adhoc(tc, "Failed to stat file: %s", uv_strerror(req.result));
@@ -60,14 +73,14 @@ MVMint64 MVM_file_stat(MVMThreadContext *tc, MVMString *filename, MVMint64 statu
         case MVM_STAT_ISDEV: {
             const int mode = file_info(tc, filename, use_lstat).st_mode;
 #ifdef _WIN32
-            r = mode & S_IFMT == S_IFCHR;
+            r = (mode & S_IFMT) == S_IFCHR;
 #else
             r = (mode & S_IFMT) == S_IFCHR || (mode & S_IFMT) == S_IFBLK;
 #endif
             break;
         }
 
-        case MVM_STAT_CREATETIME:         r = file_info(tc, filename, use_lstat).st_ctim.tv_sec; break;
+        case MVM_STAT_CREATETIME:         r = file_info(tc, filename, use_lstat).st_birthtim.tv_sec; break;
 
         case MVM_STAT_ACCESSTIME:         r = file_info(tc, filename, use_lstat).st_atim.tv_sec; break;
 
@@ -103,81 +116,36 @@ MVMint64 MVM_file_stat(MVMThreadContext *tc, MVMString *filename, MVMint64 statu
     return r;
 }
 
-/* copy a file from one to another */
-void MVM_file_copy(MVMThreadContext *tc, MVMString *src, MVMString * dest) {
-    /* TODO: on Windows we can use the CopyFile API, which is probaly
-       more efficient, not to mention easier to use. */
-    uv_fs_t req;
-    char * a, * b;
-    uv_file in_fd = -1, out_fd = -1;
-    MVMuint64 size, offset;
+MVMnum64 MVM_file_time(MVMThreadContext *tc, MVMString *filename, MVMint64 status, MVMint32 use_lstat) {
+    uv_stat_t statbuf = file_info(tc, filename, use_lstat);
+    uv_timespec_t ts;
 
-    a = MVM_string_utf8_c8_encode_C_string(tc, src);
-    b = MVM_string_utf8_c8_encode_C_string(tc, dest);
-
-    /* If the file cannot be stat(), there is little point in going any further. */
-    if (uv_fs_stat(tc->loop, &req, a, NULL) < 0)
-        goto failure;
-    size = req.statbuf.st_size;
-
-    in_fd = uv_fs_open(tc->loop, &req, (const char *)a, O_RDONLY, 0, NULL);
-    if (in_fd < 0) {
-        goto failure;
+    switch(status) {
+        case MVM_STAT_CREATETIME: ts = statbuf.st_birthtim; break;
+        case MVM_STAT_MODIFYTIME: ts = statbuf.st_mtim; break;
+        case MVM_STAT_ACCESSTIME: ts = statbuf.st_atim; break;
+        case MVM_STAT_CHANGETIME: ts = statbuf.st_ctim; break;
+        default: return -1;
     }
 
-    out_fd = uv_fs_open(tc->loop, &req, (const char *)b, O_WRONLY | O_CREAT | O_TRUNC, DEFAULT_MODE, NULL);
-    if (out_fd < 0) {
-        goto failure;
-    }
-
-    offset = 0;
-    do {
-        /* sendfile() traditionally takes offset as a pointer argument
-         * used a both input and output. libuv deviates by making
-         * offset an integer and returning the number of bytes
-         * sent. So it is necessary to add these explicitly. */
-        MVMint64 sent = uv_fs_sendfile(tc->loop, &req, out_fd, in_fd, offset, size - offset, NULL);
-        if (sent < 0) {
-            goto failure;
-        }
-        offset += sent;
-    } while (offset < size);
-
-    /* Cleanup */
-    if(uv_fs_close(tc->loop, &req, in_fd, NULL) < 0) {
-        goto failure;
-    }
-    in_fd = -1;
-
-    if (uv_fs_close(tc->loop, &req, out_fd, NULL) < 0) {
-        goto failure;
-    }
-
-    MVM_free(b);
-    MVM_free(a);
-    return;
-
- failure: {
-        /* First get the error, since it may be overwritten further on. */
-        const char * error = uv_strerror(req.result);
-        /* Basic premise: dealing with all failure cases is hard.
-         * So to simplify, a and b are allocated in all conditions.
-         * Also to simplify, in_fd are nonnegative if open, negative
-         * otherwise. */
-        MVM_free(b);
-        MVM_free(a);
-        /* If any of these fail there is nothing
-         * further to do, since we're already failing */
-        if (in_fd >= 0)
-            uv_fs_close(tc->loop, &req, in_fd, NULL);
-        if (out_fd >= 0)
-            uv_fs_close(tc->loop, &req, out_fd, NULL);
-        /* This function only throws adhoc errors, so the message is for
-         * progammer eyes only */
-        MVM_exception_throw_adhoc(tc, "Failed to copy file: %s", error);
-    }
+    return ts.tv_sec + 1e-9 * (MVMnum64)ts.tv_nsec;
 }
 
+/* copy a file from one to another */
+void MVM_file_copy(MVMThreadContext *tc, MVMString *src, MVMString * dest) {
+    char * const a = MVM_string_utf8_c8_encode_C_string(tc, src);
+    char * const b = MVM_string_utf8_c8_encode_C_string(tc, dest);
+    uv_fs_t req;
+
+    if(uv_fs_copyfile(NULL, &req, a, b, 0, NULL) < 0) {
+        MVM_free(a);
+        MVM_free(b);
+        MVM_exception_throw_adhoc(tc, "Failed to copy file: %s", uv_strerror(req.result));
+    }
+
+    MVM_free(a);
+    MVM_free(b);
+}
 
 /* rename one file to another. */
 void MVM_file_rename(MVMThreadContext *tc, MVMString *src, MVMString *dest) {
@@ -185,7 +153,7 @@ void MVM_file_rename(MVMThreadContext *tc, MVMString *src, MVMString *dest) {
     char * const b = MVM_string_utf8_c8_encode_C_string(tc, dest);
     uv_fs_t req;
 
-    if(uv_fs_rename(tc->loop, &req, a, b, NULL) < 0 ) {
+    if(uv_fs_rename(NULL, &req, a, b, NULL) < 0 ) {
         MVM_free(a);
         MVM_free(b);
         MVM_exception_throw_adhoc(tc, "Failed to rename file: %s", uv_strerror(req.result));
@@ -208,7 +176,7 @@ void MVM_file_delete(MVMThreadContext *tc, MVMString *f) {
     }
 
 #else
-    const int r = uv_fs_unlink(tc->loop, &req, a, NULL);
+    const int r = uv_fs_unlink(NULL, &req, a, NULL);
 
     if( r < 0 && r != UV_ENOENT) {
         MVM_free(a);
@@ -223,7 +191,7 @@ void MVM_file_chmod(MVMThreadContext *tc, MVMString *f, MVMint64 flag) {
     char * const a = MVM_string_utf8_c8_encode_C_string(tc, f);
     uv_fs_t req;
 
-    if(uv_fs_chmod(tc->loop, &req, a, flag, NULL) < 0 ) {
+    if(uv_fs_chmod(NULL, &req, a, flag, NULL) < 0 ) {
         MVM_free(a);
         MVM_exception_throw_adhoc(tc, "Failed to set permissions on path: %s", uv_strerror(req.result));
     }
@@ -235,8 +203,8 @@ MVMint64 MVM_file_exists(MVMThreadContext *tc, MVMString *f, MVMint32 use_lstat)
     uv_fs_t req;
     char * const a = MVM_string_utf8_c8_encode_C_string(tc, f);
     const MVMint64 result = (use_lstat
-      ? uv_fs_lstat(tc->loop, &req, a, NULL)
-      :  uv_fs_stat(tc->loop, &req, a, NULL)
+      ? uv_fs_lstat(NULL, &req, a, NULL)
+      :  uv_fs_stat(NULL, &req, a, NULL)
     ) < 0 ? 0 : 1;
 
     MVM_free(a);
@@ -247,10 +215,11 @@ MVMint64 MVM_file_exists(MVMThreadContext *tc, MVMString *f, MVMint32 use_lstat)
 #ifdef _WIN32
 #define FILE_IS(name, rwx) \
     MVMint64 MVM_file_is ## name (MVMThreadContext *tc, MVMString *filename, MVMint32 use_lstat) { \
-        if (!MVM_file_exists(tc, filename, use_lstat)) \
+        uv_stat_t statbuf; \
+        if (file_info_with_error(tc, &statbuf, filename, use_lstat) < 0) { \
             return 0; \
+        } \
         else { \
-            uv_stat_t statbuf = file_info(tc, filename, use_lstat); \
             MVMint64 r = (statbuf.st_mode & S_I ## rwx ); \
             return r ? 1 : 0; \
         } \
@@ -258,105 +227,95 @@ MVMint64 MVM_file_exists(MVMThreadContext *tc, MVMString *f, MVMint32 use_lstat)
 FILE_IS(readable, READ)
 FILE_IS(writable, WRITE)
 MVMint64 MVM_file_isexecutable(MVMThreadContext *tc, MVMString *filename, MVMint32 use_lstat) {
-    if (!MVM_file_exists(tc, filename, use_lstat))
+    MVMint64 r = 0;
+    uv_stat_t statbuf;
+    if (file_info_with_error(tc, &statbuf, filename, use_lstat) < 0)
         return 0;
+    else if ((statbuf.st_mode & S_IFMT) == S_IFDIR)
+        return 1;
     else {
-        MVMint64 r = 0;
-        uv_stat_t statbuf = file_info(tc, filename, use_lstat);
-        if ((statbuf.st_mode & S_IFMT) == S_IFDIR)
-            return 1;
-        else {
-            // true if fileext is in PATHEXT=.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC
-            MVMString *dot = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, ".");
-            MVMROOT(tc, dot, {
-                MVMint64 n = MVM_string_index_from_end(tc, filename, dot, 0);
-                if (n >= 0) {
-                    MVMString *fileext = MVM_string_substring(tc, filename, n, -1);
-                    char *ext  = MVM_string_utf8_c8_encode_C_string(tc, fileext);
-                    char *pext = getenv("PATHEXT");
-                    int plen   = strlen(pext);
-                    int i;
-                    for (i = 0; i < plen; i++) {
-                        if (0 == stricmp(ext, pext++)) {
-                             r = 1;
-                             break;
-                        }
-                    }
-                    MVM_free(ext);
-                    MVM_free(pext);
+        /* true if fileext is in PATHEXT=.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC */
+        MVMint64 n = MVM_string_index_from_end(tc, filename, tc->instance->str_consts.dot, 0);
+        if (n >= 0) {
+            MVMString *fileext = MVM_string_substring(tc, filename, n, -1);
+            char *ext  = MVM_string_utf8_c8_encode_C_string(tc, fileext);
+            char *pext = getenv("PATHEXT");
+            int plen   = strlen(pext);
+            int i;
+            for (i = 0; i < plen; i++) {
+                if (0 == stricmp(ext, pext++)) {
+                     r = 1;
+                     break;
                 }
-            });
+            }
+            MVM_free(ext);
+            MVM_free(pext);
         }
-        return r;
     }
+    return r;
 }
 #else
+
+static int are_we_group_member(MVMThreadContext *tc, gid_t group) {
+    int len;
+    gid_t *gids;
+    int res;
+    int i;
+    /* Check the user group. */
+    if (getegid() == group)
+        return 1;
+    /* Check the supplementary groups. */
+    len  = getgroups(0, NULL);
+    if (len == 0)
+        return 0;
+    gids = MVM_malloc(len * sizeof(gid_t));
+    res = getgroups(len, gids);
+    if (res < 0) {
+        MVM_free(gids);
+        MVM_exception_throw_adhoc(tc, "Failed to retrieve groups: %s", strerror(errno));
+    }
+    res = 0;
+    for (i = 0; i < len; i++) {
+        if (gids[i] == group) {
+            res = 1;
+            break;
+        }
+    }
+    MVM_free(gids);
+    return res;
+}
 #define FILE_IS(name, rwx) \
     MVMint64 MVM_file_is ## name (MVMThreadContext *tc, MVMString *filename, MVMint32 use_lstat) { \
-        if (!MVM_file_exists(tc, filename, use_lstat)) \
+        uv_stat_t statbuf; \
+        if (file_info_with_error(tc, &statbuf, filename, use_lstat) < 0) \
             return 0; \
         else { \
-            uv_stat_t statbuf = file_info(tc, filename, use_lstat); \
             MVMint64 r = (statbuf.st_mode & S_I ## rwx ## OTH) \
                       || (statbuf.st_uid == geteuid() && (statbuf.st_mode & S_I ## rwx ## USR)) \
-                      || (statbuf.st_uid == getegid() && (statbuf.st_mode & S_I ## rwx ## GRP)); \
+                      || (geteuid() == 0) \
+                      || (are_we_group_member(tc, statbuf.st_gid) && (statbuf.st_mode & S_I ## rwx ## GRP)); \
             return r ? 1 : 0; \
         } \
     }
 FILE_IS(readable, R)
 FILE_IS(writable, W)
-FILE_IS(executable, X)
-#endif
-
-/* Read all of a file into a string. */
-MVMString * MVM_file_slurp(MVMThreadContext *tc, MVMString *filename, MVMString *encoding) {
-    MVMString *mode = MVM_string_utf8_decode(tc, tc->instance->VMString, "r", 1);
-    MVMObject *oshandle = (MVMObject *)MVM_file_open_fh(tc, filename, mode);
-    MVMString *result;
-    MVM_io_set_encoding(tc, oshandle, encoding);
-    result = MVM_io_slurp(tc, oshandle);
-    MVM_io_close(tc, oshandle);
-    return result;
-}
-
-/* Writes a string to a file, overwriting it if necessary */
-void MVM_file_spew(MVMThreadContext *tc, MVMString *output, MVMString *filename, MVMString *encoding) {
-    MVMString *mode = MVM_string_utf8_decode(tc, tc->instance->VMString, "w", 1);
-    MVMObject *fh = MVM_file_open_fh(tc, filename, mode);
-    MVM_io_set_encoding(tc, fh, encoding);
-    MVM_io_write_string(tc, fh, output, 0);
-    MVM_io_close(tc, fh);
-}
-
-/* return an OSHandle representing one of the standard streams */
-MVMObject * MVM_file_get_stdstream(MVMThreadContext *tc, MVMuint8 type, MVMuint8 readable) {
-    switch(uv_guess_handle(type)) {
-        case UV_TTY: {
-            uv_tty_t * const handle = MVM_malloc(sizeof(uv_tty_t));
-            uv_tty_init(tc->loop, handle, type, readable);
-#ifdef _WIN32
-            uv_stream_set_blocking((uv_stream_t *)handle, 1);
-#else
-            ((uv_stream_t *)handle)->flags = 0x80; /* UV_STREAM_BLOCKING */
-#endif
-            return MVM_io_syncstream_from_uvstream(tc, (uv_stream_t *)handle, 1);
+    MVMint64 MVM_file_isexecutable(MVMThreadContext *tc, MVMString *filename, MVMint32 use_lstat) {
+        uv_stat_t statbuf;
+        if (file_info_with_error(tc, &statbuf, filename, use_lstat) < 0)
+            return 0;
+        else {
+            MVMint64 r = (statbuf.st_mode & S_IXOTH)
+                      || (statbuf.st_uid == geteuid() && (statbuf.st_mode & S_IXUSR))
+                      || (are_we_group_member(tc, statbuf.st_gid) && (statbuf.st_mode & S_IXGRP))
+                      || (geteuid() == 0 && (statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)));
+            return r ? 1 : 0;
         }
-        case UV_FILE:
-            return MVM_file_handle_from_fd(tc, type);
-        case UV_NAMED_PIPE: {
-            uv_pipe_t * const handle = MVM_malloc(sizeof(uv_pipe_t));
-            uv_pipe_init(tc->loop, handle, 0);
-#ifdef _WIN32
-            uv_stream_set_blocking((uv_stream_t *)handle, 1);
-#else
-            ((uv_stream_t *)handle)->flags = 0x80; /* UV_STREAM_BLOCKING */
-#endif
-            uv_pipe_open(handle, type);
-            return MVM_io_syncstream_from_uvstream(tc, (uv_stream_t *)handle, 0);
-        }
-        default:
-            MVM_exception_throw_adhoc(tc, "get_stream failed, unsupported std handle");
     }
+#endif
+
+/* Get a MoarVM file handle representing one of the standard streams */
+MVMObject * MVM_file_get_stdstream(MVMThreadContext *tc, MVMint32 descriptor) {
+    return MVM_file_handle_from_fd(tc, descriptor);
 }
 
 /* Takes a filename and prepends any --libpath value we have, if it's not an
@@ -426,7 +385,7 @@ void MVM_file_link(MVMThreadContext *tc, MVMString *oldpath, MVMString *newpath)
     char * const oldpath_s = MVM_string_utf8_c8_encode_C_string(tc, oldpath);
     char * const newpath_s = MVM_string_utf8_c8_encode_C_string(tc, newpath);
 
-    if (uv_fs_link(tc->loop, &req, oldpath_s, newpath_s, NULL)) {
+    if (uv_fs_link(NULL, &req, oldpath_s, newpath_s, NULL)) {
         MVM_free(oldpath_s);
         MVM_free(newpath_s);
         MVM_exception_throw_adhoc(tc, "Failed to link file: %s", uv_strerror(req.result));
@@ -441,7 +400,7 @@ void MVM_file_symlink(MVMThreadContext *tc, MVMString *oldpath, MVMString *newpa
     char * const oldpath_s = MVM_string_utf8_c8_encode_C_string(tc, oldpath);
     char * const newpath_s = MVM_string_utf8_c8_encode_C_string(tc, newpath);
 
-    if (uv_fs_symlink(tc->loop, &req, oldpath_s, newpath_s, 0, NULL)) {
+    if (uv_fs_symlink(NULL, &req, oldpath_s, newpath_s, 0, NULL)) {
         MVM_free(oldpath_s);
         MVM_free(newpath_s);
         MVM_exception_throw_adhoc(tc, "Failed to symlink file: %s", uv_strerror(req.result));
@@ -456,7 +415,7 @@ MVMString * MVM_file_readlink(MVMThreadContext *tc, MVMString *path) {
     MVMString *result;
 
     char * const path_s = MVM_string_utf8_c8_encode_C_string(tc, path);
-    if (uv_fs_readlink(tc->loop, &req, path_s, NULL) < 0) {
+    if (uv_fs_readlink(NULL, &req, path_s, NULL) < 0) {
         MVM_free(path_s);
         MVM_exception_throw_adhoc(tc, "Failed to readlink file: %s", uv_strerror(req.result));
     }
